from mothers_h import *
from shared import *

class DictOid:
    def __init__(self,oid,d):
        self.oid=oid
        self.store=d
    def cmp(self,d):
        s=self.store
        for k,v in d.iteritems():
            if v<>s[k]:
                return False
        return True

class MotherBox(DbDa, Speaker):
    def __init__(self,builder,filter,flag=MO_NOA,fields=None,order=None,up_dict=None,distinct=False):
        """ __init__(builder,filter[,flag=MO_NOA,fields=[],order= None,up_dict=None,distinct=None]) 

        builder is a DbMother builder.
        filter is a dict or a string.
        flag could be MO_NOA,MO_DEL,MO_LOAD.
        fields: if specified, and with MO_LOAD, only that
        fields will be loaded from DB.
        up_dict is considered only with flag=MO_UP and contains
            the dict for the update.
        distinct set DISTINCT kword after SELECT
        """

        #dbda_iface_init(self,DBDA_ADAPTER)
        speaker_iface_init(self,LOGGER)

        self.order= order

        self.builder= builder
        self.table_name= tbl= builder.table_name

        self._store= []

        if flag==MO_LOAD:
            self.loadBox(filter,fields,distinct)
        elif flag==MO_DEL:
            filter= filter or ''
            self.deleteBox(filter)
        elif flag==MO_UP:
            self.updateBox(up_dict, filter)
        else:
            pass

    def __len__(self):
        return len(self._store)

    def _retrieve_oided_mothers(self, filter):
        where= filter and self._sqlBuildFilterBox(filter, prefix= True) or ''
        tbl= self.table_name
        qry= "SELECT %(tbl)s.oid from %(tbl)s %(where)s" % locals()

        res= self.mr_query(qry)
        mothers= [self.builder(oid=d['oid']) for d in res]

        return mothers

    def _retrieve_stored_mothers(self):
        mothers= [self.builder(oid=d['oid']) for d in self.getFields()]
        return mothers

    def _trigger_mothers(self, mothers, flag, when):
        if len(mothers):
            m= mothers[0]
            if not m._triggers_are_initialized() or \
                not m.get_flag_triggers(flag):
                    return

            for m in mothers:
                m.trigger(flag, when)

    def _sqlFieldsComma(self, store):
        """ _sqlFieldsComma(store) --> string

        Returns 'f_1=%(f_1)s,f_2=%(f_2), ....'
        or 'f=DEFAULT, ...' if store[f]=SQL_DEFAULT

        Ah, sure, store is a dict.
        """
        tl=[]
        for k,v in store.iteritems():
            if v==SQL_DEFAULT:
                tl.append(k+" = DEFAULT")
            else:
                tl.append(k+(" = %%(%s)s" % k))
        return ','.join(tl)


    def _sqlBuildFilterBox(self,filter,prefix=False):
        """ _sqlBuildFilterBox(self,filter, prefix= False) --> str

        Build the filter as SQL string. filter could be
        a string or a dict. 
        Note that WHERE is not included by default: you
        have to use prefix= True. By the way, if WHERE is already
        present in filter, it's not added.
        """

        s= None

        if filter is None:
            s= ''
        if isinstance(filter, str):
            s= filter

        if isinstance(filter, dict):
            if len(filter)==0:
                return ""
            tl=[(item+(" = %%(%s)s" % item)) for item in filter]
            s= " AND ".join(tl)
            s=self.mogrify(s,filter)
        
        if s is None:
            self.log_int_raise("Invalid filter type: %s" % 
                    RED(str(type(filter))))

        if not prefix or 'WHERE' in s:
                return s
        return " WHERE %s " % s 

    def loadBox(self, filter, fields= None, distinct= False):
        ms= self._retrieve_oided_mothers(filter)
        self._trigger_mothers(ms, MO_LOAD, MO_BEFORE)
        self._loadBox(filter, fields, distinct)
        self._trigger_mothers(ms, MO_LOAD, MO_AFTER)

    def _loadBox(self,filter,fields=None,distinct=False):
        """ loadBox(self,filter,fields,distinct=False) --> None

        load data from DB.
        if fields is specified, only that fields will be loaded.
        if distinct, it will be a SELECT DISTINCT
        """

        tbl=self.table_name
        self.log_debug("Mbox: loading with filter=%s",\
            YELLOW(filter),DBG_INSANE)
        
        dist = distinct and 'DISTINCT' or ''

        sel="SELECT %(dist)s %(tbl)s.oid,%(what)s FROM %(tbl)s %(where)s "

        if fields is None:
            what="%s.*" % tbl
        else:
            if not len(fields):
                self.log_int_raise("Ei, if you want to choose some "\
                                    "fields, do it correctly.")
            what=",".join(['%s.%s' % (tbl,f) for f in fields])

        if filter is not None:
            where=self._sqlBuildFilterBox(filter, prefix= True)
        else:
            where=''

        s=self._store
        qry= sel % locals()

        if self.order:
            qry_order= " ORDER BY %s"\
                % ','.join(self.order)
            qry+=qry_order

        res=self.mr_query(qry)
        for d in res:
            oid=d.pop('oid')
            s.append(DictOid(oid,d))

        self.log_debug("Loaded %s records on %s",\
                GREEN(len(s)),GREEN(tbl),DBG_NORMAL)

        self.log_debug("Action %s.", GREEN("completed"), DBG_NORMAL)

    def _deleteBoxOid(self):
        s=self._store
        if len(s)==0:
            self.log_debug("MthBox: Nothing to delete.",DBG_INSANE)
            return

        oids=["oid=%d" % r.oid for r in s]
        return " OR ".join(oids)

    def deleteBox(self,filter=None, stored= False):
        if stored:
            ms= self._retrieve_stored_mothers()
        else:
            ms= self._retrieve_oided_mothers(filter)

        self.log_debug("Before Box Triggering....", DBG_INSANE)
        self._trigger_mothers(ms, MO_DEL, MO_BEFORE)
        self._deleteBox(filter, stored)
        self.log_debug("AFter Box Triggering....", DBG_INSANE)
        self._trigger_mothers(ms, MO_DEL, MO_AFTER)

    def _deleteBox(self,filter=None, stored= False):
        """ deleteBox(self, filter= None, stored= False) --> None

        delete from DB.
        If stored is True, delete stored object.
        """

        if stored:
            if not len(self):
                self.log_debug("Box: nothing to delete",DBG_INSANE)
                return
            self.log_debug("Box: deleting stored records",DBG_INSANE)
            frecs= self.filterBox(filter, pop= True)
            oids=["oid=%d" % r.oid for r in s]
            where= " WHERE %s " % " OR ".join(oids)

        else: 
            self.log_debug("Mbox: deleting with filter=%s",\
                YELLOW(filter),DBG_INSANE)
            where= filter and self._sqlBuildFilterBox(filter, prefix= True) or ''

        tbl=self.table_name

        qry="DELETE FROM %(tbl)s %(where)s" % locals()
        self.oc_query(qry)
        self.log_debug("Records deleted.",DBG_INSANE)

        self.log_debug("Action %s.", GREEN("completed"), DBG_NORMAL)

    def updateBox(self, d, filter=None):
        ms= self._retrieve_oided_mothers(filter)
        self._trigger_mothers(ms, MO_UP, MO_BEFORE)
        self._updateBox(d, filter)
        self._trigger_mothers(ms, MO_UP, MO_AFTER)

    def _updateBox(self, d, filter=None):
        """ updateBox(self, d, filter=None): --> None

        Update a set of records.
        """

        if filter is not None:
            where= self._sqlBuildFilterBox(filter, prefix= True)
        else:
            filter= ''

        table= self.table_name
        what= self._sqlFieldsComma(d)
        what= self.mogrify(what, d)

        qry= "UPDATE %(table)s SET %(what)s %(where)s" % locals()
        qry= self.mogrify(qry, d)

        self.oc_query(qry)
        self.log_debug("Box: Updatetd records.", DBG_INSANE)

        self.log_debug("Action %s.", GREEN("completed"), DBG_NORMAL)

    def filterBox(self,filter,inverse=False, pop= False):
        """ filterBox(self,filter,pop=False) --> set

        for each stored object, if filter apply on it,
        return it.
        If inverse, filter is inverted.
        Results are pop'd from store.
        """

        # Matching ...
        popped= []
        for r in self._store:
            ok= r.cmp(filter)
            if ok <> inverse:
                popped.append(r)

        if pop:
            for p in popped:
                self._store.remove(p)

        self.log_debug("Filtered %s records",\
                    GREEN(len(popped)),DBG_INSANE)

        return popped

    def getFields(self,filter=None,inverse=False,flag_obj=False,fields=None,cdict=None, pop= False):
        """ getFields(self,filter=None,inverse=False,flag_obj=False,fields=[],cdict=None,pop=False) --> list 

        Returns the stored objects.
        If filter is specified, exclude objects filtered (or, if inverse is True,
        exclude objects that don't match the filter).
        if flag_obj, returns a list of Mothers Objects (fields and cdict are ignored).
        If flag_obj is False, returns a list of dict.
        In such case, if fields is specified, returns only these fields.
        If cdict, each dict in the returned list will be updated with cdict.
        If pop, results will be deleted from local store (speeds up next calls).
        """

        if filter:
            s= self.filterBox(filter, inverse, pop) 
        else:
            s= self._store
            if pop:
                self._store= []

        res=[]
        for r in s:
            if flag_obj:
                oid= r.oid
                n= self.builder(r.store, oid= oid)

            else:
                d= pop and r.store or r.store.copy()
                if fields is not None:
                    for k in d.keys():
                        if k not in fields:
                            d.pop(k)
                if cdict:
                    d.update(cdict)
                n= d
            res.append(n)
        return res
            
