#!/usr/bin/python

import sys
import copy
import getopt
import os

from speaker import *
from dbda import *

from metaqry import *
from mothers import getMotherObj, MO_SAVE

from shared import DEFAULT_LOGGER

Logger= DEFAULT_LOGGER
Logger.set_log_level(0)
Db=DbDa()


# Global
STRUCT_FILE= 'dbstruct.py'
RELS_FILE= 'rels.py'
_DROP_TABLES_SPEED_UP= True

_STDERR= sys.stderr
_STDOUT= sys.stdout
_STDERR_BKP= None

_QUIET= False
def quiet():
    global _QUIET
    _QUIET= True

_NOFILES= False
def nofiles():
    global _NOFILES
    _NOFILES= True

RED_D=RED("D")
RED_ALERT=RED("Error")
GREEN_DONE= GREEN("Done\n")

#
## Print
#

def loginfo(s):
    _STDERR.write(s)
    _STDERR.flush()

def stdoutinfo(s):
    _STDOUT.write(s)
    _STDOUT.flush()

def outlog(s):
    if _QUIET:
        return
    loginfo(s)

def stderr_to(f):
    global _STDERR_BKP, _STDERR
    _STDERR_BKP= _STDERR
    try:
        _STDERR= open(f,'w')
    except:
        _STDERR= _STDERR_BKP
        return -1

def stderr_back():
    global _STDERR
    _STDERR.close()
    _STDERR= _STDERR_BKP

#        
##      
### Twirl 
##      Inspired by portage (Gentoo)     
#     

twirl_sequence = "/-\\|/-\\|/-\\|/-\\|\\-/|\\-/|\\-/|\\-/|"
twirl_spinops=0
twirl_len=len(twirl_sequence)


def twirl():
    global twirl_sequence
    global twirl_spinops
    global twirl_len
    twirl_spinops=(twirl_spinops + 1) % twirl_len
    outlog("\b\b "+ twirl_sequence[twirl_spinops])
    return True

#
##
###
##
#

class God:
    def __init__(self, devil):
        self.children= set()
        self.orphaned= set()
        self.devil= devil
        self.disordered= set()

    def clean(self):
        self.children=set()
        self.orphaned= set()
        self.disordered= set()
        self.devil.clean()

    def load_tables(self):
        res= Db.mr_query(TBL_QUERY)
        tables= [twirl() and d['table_name'] for d in res]
        return tables

    def fill_God(self):
        outlog( "\tLoading Mother World...   ")
        sys.stdout.flush()
        twirl()
        tables= self.load_tables()
        twirl()
        for tbl in tables:
            twirl()
            orphan= Tbl(tbl)
            self.orphaned.add(orphan)
            self.disordered.add(orphan)
        outlog("\b\b %s" % GREEN_DONE)

    def add_child(self, c):
        self.children.add(c)

    def find_fathers(self, tbl):
        res=set()
        for child in self.children:
            res |= child.find_fathers(tbl)
        return res

    def build_God(self):
        outlog("\tBuilding God Picture...   ")
        sys.stdout.flush()

        # first of all, get prophetes

        oped= self.orphaned
        twirl()
        _oped= oped.copy()

        temp_l= [orphan for orphan in oped \
                if orphan.is_god_child()]

        twirl()
        temp_set=set(temp_l)
        self.children |= temp_set
        oped-= temp_set

        l= len(oped)
        while l:
            twirl()
            temp_set.clear()
            for o in oped:
                twirl()
                o_fathers= o.get_unlinked_fathers()
                for f in o_fathers:
                    twirl()
                    s=self.find_fathers(f)
                    for pf in s:
                        twirl()
                        o.add_father(pf)

                if o.is_parented():
                    temp_set.add(o)

            twirl()
            oped-=temp_set
            l= len(oped)

        devil= self.devil
        for o in _oped:
            twirl()
            if not len(o.children):
                devil.add_father(o)
        outlog( "\b\b %s" % GREEN_DONE)



class Devil:
    def __init__(self):
        self.fathers=set()

    def add_father(self, f):
        self.fathers.add(f)

    def pop_damned(self):
        res=[]
        second_level= set()
        for f in self.fathers:
            second_level |= f.fathers
            for ff in f.get_fathers():
                f.del_father(ff)
            res.append(f)

        self.fathers.clear()
        for f in second_level:
            if f.is_devil_father():
                self.add_father(f)

        return res
    
    def clean(self):
        self.fathers=set()


class Tbl:
    def __init__(self, name):
        self.name= name
        self.fields= []
        self.pkeys= []
        self.fathers_map= [] # (mykey, tbl, key)
        self.unlinked_fathers= set()
        self.fathers= set()
        self.children= set()

        self.load_keys()
        self.load_pkeys()
        self.load_fkeys()

    def get_father_map(self, tbl):
        for a,b,c in self.fathers_map:
            if b==tbl:
                return a,c

    def is_god_child(self):
        if len(self.fathers_map):
            return False
        return True
    
    def is_devil_father(self):
        if len(self.children):
            return False
        return True

    def load_pkeys(self):
        qry= TBL_PKEYS_QUERY % self.name
        res= Db.mr_query(qry)
        self.pkeys= [d['column_name'] for d in res]

    def load_fkeys(self):
        mykeys=[]
        qry= TBL_FKEYS_QUERY % self.name
        res= Db.mr_query(qry)
        for d in res:
            buf=d['tgargs']
            s= str(buf)
            l= s.split("\x00")
            mykey= l[4]
            if mykey in mykeys:
                continue
            tbl= l[2]
            key= l[5]
            self.fathers_map.append((mykey, tbl, key))
            self.unlinked_fathers.add(tbl)
            mykeys.append(mykey)

    def load_keys(self):
        qry= TBL_KEYS_QUERY % self.name
        res= Db.mr_query(qry)
        self.keys= [d['column_name'] for d in res]

    def add_father(self, tbl_obj):
        tbl= tbl_obj.name
        self.unlinked_fathers.remove(tbl)
        self.fathers.add(tbl_obj)
        tbl_obj.add_child(self)

    def del_father(self, tbl_obj):
        self.fathers.remove(tbl_obj)
        tbl_obj.children.remove(self)

    def add_child(self, tbl_obj):
        self.children.add(tbl_obj)

    def is_parented(self):
        return len(self.unlinked_fathers)==0

    def get_fathers(self):
        return self.fathers.copy()

    def get_unlinked_fathers(self):
        return list(self.unlinked_fathers)

    def find_fathers(self, tbl):
        res=set()
        if self.name==tbl:
            res.add(self)
        for child in self.children:
            res |= child.find_fathers(tbl)
        return res

class MotherWorld:
    def __init__(self, scripts= []):
        # Devil first, and after that, God.
        devil= Devil()
        self.devil= devil
        # Yes, God needs Devil
        self.god= God(devil)
        self.world_created= False
        self.scripts= scripts
        self.symbols_files= []

    def add_symbols_file(self, f):
        if not f.endswith('.py'):
            loginfo("Symbols file has to be a python file. "\
                    "%s won't be loaded.\n" % RED(f))
            return -1
        
        self.symbols_files.append(f[:-3])
        return 0

    def load_symbols_file(self):
        f= self.symbols_files.pop(0)
        self._load_symbols_files(f)

    def _load_symbols_files(self, f):
        p= os.path
        pyfil= "%s.py" % f
        outlog("Processing symbol file %s ...\n" % GREEN(pyfil))

        cfil= p.abspath(f)
        bfil= p.basename(f)

        index= cfil.index(bfil)
        if index:
            dir= cfil[:index]
            sys.path= [dir]+ sys.path

        try:
            mod=__import__(bfil)
        except:
            loginfo("\n\tError: unable to import %s. "\
                    "Symbol file won't be processed.\n" % RED(bfil))
            return

        tbi= mod.RECORDS_TO_BE_INSERTED

        tcount=0
        for t,dl in tbi:
            count= 0
            outlog("\tWorking on table %s... " % GREEN(t))
            Db.beginTrans()
            try:
                Db.oc_query("DELETE FROM %s" % t)
                for d in dl:
                    m= getMotherObj(t, d, MO_SAVE, dbda= Db, logger= Logger)
                    count+= 1
                Db.commit()

            except Exception, s:
                loginfo("Error: %s; " % RED(s))
                Db.rollback()
                count= 0

            tcount+= count
            outlog("inserted %s records\n" % GREEN(count))

        outlog("Inserted %s records.\n" % GREEN(tcount))

    def create_world(self):

        if self.world_created:
            return

        outlog( "Creating Mother World...\n")
        god= self.god
        god.clean()
        god.fill_God()
        god.build_God()
        outlog("\n")
        self.world_created= True

    def pop_damned(self):
        return self.devil.pop_damned()

    def drop_all_tables(self):

        outlog("Deleting all tables...")
        if _DROP_TABLES_SPEED_UP:
            tbls= self.god.load_tables()
            outlog("\n")
            count=0
            for t in tbls:
                Db.oc_query("DROP TABLE %s cascade" % t)
                outlog( "\t%s %s\n" % (RED_D, RED(t)))
                count+=1
            self.world_created= False
            outlog("Dropped %s\n" % RED(count))
            return

        self.create_world()

        count=0
        r= self.pop_damned()
        while r:
            for tbl_obj in r:
                tbl= tbl_obj.name
                Db.oc_query(TBL_DEL_QUERY % tbl)
                outlog( "\t%s %s\n" % (RED_D, RED(tbl)))
                count+=1
            r= self.pop_damned()

        self.world_created= False
        outlog("Dropped %s\n" % RED(count))

    def drop_all_records(self):
        self.create_world()

        world= copy.copy(self)

        r= world.pop_damned()
        while r:
            for tbl_obj in r:
                tbl= tbl_obj.name
                Db.oc_query("DELETE FROM %s" % tbl)
                outlog( "\t%s %s.%s\n" % (RED_D, RED(tbl), GREEN("*")))
            r= self.pop_damned()

        del world

    def add_script(self, s):
        self.scripts.append(s)

    def exec_script(self):
        s=self.scripts.pop(0)
        self._exec_script(s)

    def _exec_script(self, script):
        outlog("Executing script %s... " % YELLOW(script))
        try:
            fil=open(script)
            buf= fil.read()
            fil.close()
        except:
            outlog("%s: unable to read script.\n" % RED_ALERT)
            return

        try:
            Db.oc_query(buf)
        except Exception, s:
            loginfo("%s: %s" % (RED_ALERT, s))
            sys.exit(1)

        outlog(GREEN_DONE)

    def create_db_struct(self):

        self.create_world()
        outlog("Creating db structure... ")

        buf="# Created by dbmapper.\n"\
                "TABLE_FIELDS_DICT={}\nTABLE_PKEYS_DICT={}\n\n"

        for tbl_obj in self.god.disordered:
            tbl= tbl_obj.name
            TBL= "TABLE_%s" % tbl.upper()
            buf+=("%s= '%s'\n" % (TBL, tbl))
            buf+=("TABLE_FIELDS_DICT[%s]= %s\n" % (TBL, list(tbl_obj.keys)))
            buf+=("TABLE_PKEYS_DICT[%s]= %s\n\n" % (TBL, list(tbl_obj.pkeys)))

        if _NOFILES:
            stdoutinfo(buf)
        else:
            fil=open(STRUCT_FILE,'w')
            fil.write(buf)
            fil.close()

        outlog("%s\n" % GREEN_DONE)


    def _depdict_buf(self, obj):
        c=0
        tbl= obj.name
        TBL= "TABLE_%s" % tbl.upper()
        buf=("    %s:{\n" % TBL)

        for child in obj.children:
            c+=1
            mykey, fkey= child.get_father_map(tbl)
            CTBL= "TABLE_%s" % child.name.upper()

            buf+=("            "\
                    "%s:{'%s':'%s'},\n" %\
                    (CTBL, fkey, mykey))

        buf+=("        },\n")

        return c, buf
    
    def _relation_buf(self, obj):
        c=0

        tables= []
        for f in obj.fathers:
            c+=1
            tables.append("TABLE_%s" % f.name.upper())
        buf="    _freeze(%s):\\\n"\
                "                    "\
                "TABLE_%s,\n" % (','.join(tables), obj.name.upper())
        return c, buf

    def create_db_rels(self):

        self.create_world()
        outlog("Creating relations rules...\n")

        count_fathers= 0
        count_children= 0
        count_rels= 0
        count_frels= 0

        dbuf=("# created by dbmapper\n"\
                "from dbstruct import *\n\n"\
                "def _freeze(*args):\n"\
                "    return frozenset(args)\n\n"\
                "TABLE_CHILDS_DICT={\n")

        rbuf=("TABLE_RELATIONS_DICT={\n")

        for tbl_obj in self.god.disordered:

            if len(tbl_obj.fathers)>1:
                rf, buf= self._relation_buf(tbl_obj)
                count_rels+= 1
                count_frels+= rf
                rbuf+= buf

            if tbl_obj.is_devil_father():
                continue

            cc, buf= self._depdict_buf(tbl_obj)
            count_fathers+= 1
            count_children+= cc
            dbuf+= buf
        
        dbuf+=("}\n\n")
        rbuf+=("}\n")
        buf= dbuf+ rbuf


        if _NOFILES:
            stdoutinfo(buf)
        else:
            fil=open(RELS_FILE,'w')
            fil.write(buf)
            fil.close()
            outlog("\tWrited %s dependencies dict(s) for %s child(ren).\n" %\
                (GREEN(count_fathers),GREEN(count_children)))
            outlog("\tWrited %s relation dict(s) for %s father(s).\n" %\
                (GREEN(count_rels),GREEN(count_frels)))

        outlog("\n")

def testdb():
    try:
        Db._connect()
        outlog("\tDatabase Connection is %s.\n" % GREEN('OK'))
    except Exception, s:
        outlog("\tDatabase Connection %s: %s\n" % (RED('BROKEN'),s))

def fill_symbols_usage():
    loginfo(
            """
  dbmapper -f file

  This option is used to create at the same time code symbols and
  db records, and to sync them.
  If you use a table that contains some global, static values, 
  this should be a good option.

  For example, consider the following tables:
  
   CREATE TABLE color (color int, name text);
   CREATE TABLE colored_object (color int, ..., foreign key (color) references color(color) )
  
  Suppose that the color table is static: you have YELLOW=1, MARBLE=2, etc.

  So, probably you'll define symbols in your code, to :

    YELLOW= 1
    MARBLE= 2

  To be sure that DB records and symbols will be always synced, you can
  build the following python file:

    ### BEGIN FILE ###

        YELLOW= 1
        MARBLE= 2

        RECORDS_COLOR=[
                    dict(color= YELLOW, name= 'Yellow'),
                    dict(color= MARBLE, name= 'Yellow')
                    ]
    
        RECORDS_TO_BE_INSERTED=(
                # tuples: (table_name, list_of_records)
                ('color', RECORDS_COLOR)
                )

    ### END FILE ###

  If this file will be used with option -f, the records specified will be
  inserted on DB and, at the same time, you can import the file and use
  the symbols defined here.

  In this way, you will be sure that symbols and records will be synced.\n\n""")
    sys.exit(1)

def usage():
    loginfo(
    """
  Usage: dbmapper OPTIONS
  dbmapper is the central tool for Mother configuration.

  OPTIONS:
      -q (--quiet)                  don't love loqaucity?
      -o (--output)                 for options s,r: if specified, don't write files,
                                      print them to standard output
      -d (--drop-tables)            drop all tables in db
      -E (--empty-tables)           drop all records in all tables
      -e (--exec-script= script)    execute the SQL script `script`
      -s (--dbstructure)            create dbstruct.py, or, if option `o` is specified,
                                      print the file contents on stdout
      -r (--relations)              create rels.py, or, if option `o` is specified,
                                      print the file contents on stdout
      -t (--testdb)                 test db connection
      -h (--help)                   help me please
      -f (--fill-symbols file)      fill db sybols. `-f help` to know more\n\n""")

    sys.exit(1)

def main():

    opts= "qodEe:srthf:"
    opts_long=[
            'quiet',
            'output',
            'drop-tables',
            'empty-tables',
            'exec-script',
            'dbstructure',
            'relations',
            'testdb',
            'help',
            'fill-symbols'
            ]

    try:
        ops, args= getopt.getopt(sys.argv[1:], opts, opts_long)
        assert len(ops)
    except:
        usage()

    w= MotherWorld()
    functors=[]

    for o,a in ops:
        if o in ['-q','--quiet']:
            quiet()
        elif o in ['-o','--output']:
            nofiles()
        elif o in ['-d','--drop-tables']:
            functors.append(w.drop_all_tables)
        elif o in ['-E','--empty-tables']:
            functors.append(w.drop_all_records)
        elif o in ['-e','--exec-script']:
            w.add_script(a)
            functors.append(w.exec_script)
        elif o in ['-s','--dbstructure']:
            functors.append(w.create_db_struct)
        elif o in ['-r','--relations']:
            functors.append(w.create_db_rels)
        elif o in ['-t', '--testdb']:
            testdb()
        elif o in ['-h','--help']:
            usage()
        elif o in ['-f','--fill-symbols']:
            if a=='help':
                fill_symbols_usage()
            res= w.add_symbols_file(a)
            if not res:
                functors.append(w.load_symbols_file)
        else:
            loginfo("Invalid option '%s'\n" % o)

    for f in functors:
        f()
 
if __name__=='__main__':
    main()
